import inference as ifr
import arma as arma
import output as op
import tests as tst
import distributions as dst
import numpy as np
from math import exp, sqrt, log, tanh
from scipy import optimize
import copy
import matplotlib.pyplot as plt
import matplotlib.mlab as mlab
import seaborn as sns
import numdifftools as nd
import covariances as cov
import pandas as pd

class TSM(object):
	""" TSM PARENT CLASS

	Contains general time series methods to be inherited by models.

	Parameters
	----------
	model_type : str
		The type of model (e.g. 'ARIMA', 'GARCH')
	"""

	def __init__(self,model_type):

		# Holding variables for model output
		self.params = []
		self.ses = []
		self.ihessian = []
		self.chains = []
		self.predictions = []

		# Holds parameter descriptions and model type
		self.param_desc = []
		self.model_type = model_type

	def posterior(self,beta):
		""" Returns negative log posterior

		Parameters
		----------
		beta : np.ndarray
			Contains untransformed starting values for parameters

		Returns
		----------
		Negative log posterior
		"""

		post = self.likelihood(beta)
		for k in range(self.param_no):
			post += -self.param_desc[k]['prior'].logpdf(beta[k])
		return post

	def adjust_prior(self,index,prior):
		""" Adjusts priors for the parameters

		Parameters
		----------
		index : int or list[int]
			Which parameter index/indices to be altered

		prior : PRIOR object
			Which prior distribution? E.g. Normal(0,1)

		Returns
		----------
		None (changes priors in self.param_desc)
		"""

		if isinstance(index, list):
			for item in index:
				if item < 0 or item > (self.param_no-1) or not isinstance(item, int):
					raise Exception("Oops - the parameter index " + str(item) + " you have entered is invalid!")
				else:
					self.param_desc[item]['prior'] = prior				
		else:
			if index < 0 or index > (self.param_no-1) or not isinstance(index, int):
				raise Exception("Oops - the parameter index " + str(index) + " you have entered is invalid!")
			else:
				self.param_desc[index]['prior'] = prior

	# Wrapper function for inference options
	def fit(self,method=None,printer=True,nsims=100000,cov_matrix=None,step=0.001,iterations=30000,**kwargs):
		""" Fits a model

		Parameters
		----------
		method : str
			A fitting method (e.g 'MLE')

		printer : Boolean
			Whether to print results or not

		nsims : int (optional)
			How many simulations if an MCMC method

		cov_matrix : None or np.ndarray
			Option to include covariance matrix for M-H MCMC

		step : float
			Step size for BBVI method

		iterations = int
			How many iterations for BBVI

		Returns
		----------
		None (stores fit information)
		"""

		if method is None:
			method = self.default_method

		if method not in self.supported_methods:
			raise ValueError("Method not supported!")

		if method == 'MLE':
			self.optimize_fit(self.likelihood,printer)
		elif method == 'MAP':
			self.optimize_fit(self.posterior,printer)	
		elif method == 'M-H':
			self.mcmc_fit(nsims=nsims,method=method,cov_matrix=cov_matrix)
		elif method == "Laplace":
			self.laplace_fit(self.posterior,printer) 
		elif method == "BBVI":
			self.bbvi_fit(self.posterior,printer,step=step,iterations=iterations)
		elif method == "OLS":
			self.ols_fit(printer)			
		else:
			raise ValueError("Method not recognized!")

	def list_priors(self):
		""" Lists the current prior specification

		Returns
		----------
		None (prints out prior information)
		"""

		prior_list = []
		prior_desc = []
		param_trans = []

		for i in range(self.param_no):
			param_trans.append(self.param_desc[i]['prior'].transform_name)
			x = self.param_desc[i]['prior']
			if isinstance(x, ifr.Normal):
				prior_list.append('Normal')
				prior_desc.append('mu0: ' + str(self.param_desc[i]['prior'].mu0) + ', sigma0: ' + str(self.param_desc[i]['prior'].sigma0))
			elif isinstance(x, ifr.InverseGamma):
				prior_list.append('Inverse Gamma')
				prior_desc.append('alpha: ' + str(self.param_desc[i]['prior'].alpha) + ', beta: ' + str(self.param_desc[i]['prior'].beta))
			elif isinstance(x, ifr.Uniform):
				prior_list.append('Uniform')
				prior_desc.append('n/a (non-informative)')
			else:
				raise Exception("Error - prior distribution not detected!")

		data = []

		for i in range(self.param_no):
			data.append({'param_index': len(data), 'param_name': self.param_desc[i]['name'], 'param_prior': prior_list[i], 'param_prior_params': prior_desc[i], 'param_trans': param_trans[i]})

		fmt = [
			('Index','param_index',6),		
			('Parameter','param_name',25),
			('Prior Type','param_prior',10),
			('Prior Parameters', 'param_prior_params', 25),
			('Transformation','param_trans',20)]
		
		print( op.TablePrinter(fmt, ul='=')(data) )

	def list_q(self):
		""" Lists the current approximating distributions for variational inference

		Returns
		----------
		None (lists prior specification)
		"""

		q_list = []

		for i in range(self.param_no):
			x = self.param_desc[i]['q']
			if isinstance(x, dst.Normal):
				q_list.append('Normal')
			elif isinstance(x, dst.InverseGamma):
				q_list.append('Inverse Gamma')
			elif isinstance(x, dst.Uniform):
				q_list.append('Uniform')
			else:
				raise Exception("Error - prior distribution not detected!")

		data = []

		for i in range(self.param_no):
			data.append({'param_index': len(data), 'param_name': self.param_desc[i]['name'], 'param_q': q_list[i]})

		fmt = [
			('Index','param_index',6),		
			('Parameter','param_name',25),
			('q(z)','param_q',10)]
		print "Approximate distributions for mean-field variational inference"
		print ""
		print( op.TablePrinter(fmt, ul='=')(data) )

	def normal_posterior_sim(self,parameters,cov_matrix,method_name,printer):
		""" Simulates from a multivariate normal posterior

		Parameters
		----------
		parameters : np.ndarray
			Contains untransformed starting values for parameters

		cov_matrix : np.ndarray
			The inverse Hessian

		method_name : str
			Name of the estimation procedure using this function

		printer : str
			Whether to print results or not

		Returns
		----------
		None (plots the posteriors)
		"""

		chain, mean_est, median_est, upper_95_est, lower_95_est = ifr.norm_post_sim(parameters,cov_matrix)

		# Transform parameters - replace if statement with transform information directly in future version
		for k in range(len(chain)):
			if self.param_desc[k]['prior'].transform == np.exp:
				chain[k] = np.exp(chain[k])
				mean_est[k] = exp(mean_est[k])
				median_est[k] = exp(median_est[k])
				upper_95_est[k] = exp(upper_95_est[k])
				lower_95_est[k] = exp(lower_95_est[k])				
			elif self.param_desc[k]['prior'].transform == np.tanh: 
				chain[k] = np.tanh(chain[k])
				mean_est[k] = tanh(mean_est[k])
				median_est[k] = tanh(median_est[k])
				upper_95_est[k] = tanh(upper_95_est[k])
				lower_95_est[k] = tanh(lower_95_est[k])	

		self.chains = chain

		if printer is True:

			data = []

			for i in range(len(self.param_desc)):
				data.append({'param_name': self.param_desc[i]['name'], 'param_mean': round(mean_est[i],4), 'param_median': round(median_est[i],4), 'ci':  "(" + str(round(lower_95_est[i],4)) + " | " + str(round(upper_95_est[i],4)) + ")"})

			fmt = [
				('Parameter','param_name',20),
				('Median','param_median',10),
				('Mean', 'param_mean', 15),
				('95% Credibility Interval','ci',25)]

			if self.model_type == 'ARIMA':
				self.model_name = "ARIMA(" + str(self.ar) + "," + str(self.integ) + "," + str(self.ma) + ") regression"
				self.cutoff = max(self.ar,self.ma)
			elif self.model_type == 'GAS':
				self.model_name = "GAS(" + str(self.ar) + "," + str(self.integ) + "," + str(self.sc) + ") regression"
				self.cutoff = max(self.ar,self.sc)
			elif self.model_type == 'VAR':
				self.model_name = "VAR(" + str(self.lags) + ") regression"
				self.cutoff = self.lags			
			elif self.model_type in ['EGARCH', 'GARCH']:
				self.model_name = self.model_type + "(" + str(self.p) + "," + str(self.q) + ") regression"
				self.cutoff = max(self.p,self.q)	
			elif self.model_type == 'GPNARX':
				self.model_name = "GP-NARX(" + str(self.ar) + ") regression"
				self.cutoff = self.max_lag	

			print self.model_name 
			print "=================="
			print "Method: " + method_name + " Approximation"
			print "Number of observations: " + str(len(self.data)-self.cutoff)				
			print "Unnormalized Log Posterior: " + str(round(-self.posterior(mean_est),4))
			print ""
			print( op.TablePrinter(fmt, ul='=')(data) )
		
		# Plot densities
		for j in range(len(self.params)):
			fig = plt.figure()
			a = sns.distplot(chain[j], rug=False, hist=False, label=self.param_desc[j]['name'])
			a.set_title(self.param_desc[j]['name'])
			a.set_ylabel('Density')
				
		sns.plt.show()		

	def laplace_fit(self,obj_type,printer=True):
		""" Performs a Laplace approximation to the posterior

		Parameters
		----------
		obj_type : method
			Whether a likelihood or a posterior

		printer : Boolean
			Whether to print results or not

		Returns
		----------
		None (plots posterior)
		"""

		# Get Mode and Inverse Hessian information
		self.fit(method='MAP',printer=False)

		if len(self.ses) == 0: # no errors, no Laplace party
			raise Exception("No Hessian information - Laplace approximation cannot be performed")
		else:
			self.normal_posterior_sim(self.params,self.ihessian,"Laplace",printer)

	def optimize_fit(self,obj_type,printer=True):
		""" Performs optimization of an objective function

		Parameters
		----------
		obj_type : method
			Whether a likelihood or a posterior

		printer : Boolean
			Whether to print results or not

		Returns
		----------
		None (stores parameters)
		"""

		# Starting parameters
		phi = np.zeros(self.param_no)

		# Starting values
		if self.model_type == 'GAS':
			if self.dist in ['Laplace','Normal']:
				phi[0] = np.mean(self.data)
			elif self.dist == 'Poisson':
				phi[0] = np.log(np.mean(self.data))
		elif self.model_type == 'VAR':
			phi = self.create_B_direct().flatten()
			cov = self.estimate_cov()

			# Inelegant - needs refactoring
			for i in range(self.ylen):
				for k in range(self.ylen):
					if i == k:
						phi = np.append(phi,self.param_desc[len(phi)]['prior'].itransform(cov[i,k]))
					elif i > k:
						phi = np.append(phi,self.param_desc[len(phi)]['prior'].itransform(cov[i,k]))			
		elif self.model_type == 'GARCH':
			phi = np.ones(self.param_no)*0.00001
			phi[0] = self.param_desc[0]['prior'].itransform(np.mean(np.power(self.data,2)))
		elif self.model_type == 'EGARCH':
			phi[0] = self.param_desc[0]['prior'].itransform(np.log(np.mean(np.power(self.data,2))))
		elif self.model_type == 'GPNARX':
			phi = np.ones(self.param_no)*-1.0
			arma_start = arma.ARIMA(self.data,ar=self.ar,ma=0,integ=self.integ)
			arma_start.fit(printer=False)
			phi[0] = arma_start.params[len(arma_start.params)-1]			

		# Optimize using L-BFGS-B
		p = optimize.minimize(obj_type,phi,method='L-BFGS-B')
		self.params = copy.deepcopy(p.x)

		# Vector for transformed parameters
		t_params = copy.deepcopy(p.x)			

		# Check that matrix is non-singular; act accordingly
		try:
			self.ihessian = np.linalg.inv(nd.Hessian(obj_type)(self.params))
			self.ses = np.abs(np.diag(self.ihessian))**0.5

			t_p_std = copy.deepcopy(self.ses) # vector for transformed standard errors
			
			# Create transformed variables
			for k in range(len(t_params)-self.param_hide):
				z_temp = (p.x[k]/float(self.ses[k]))
				t_params[k] = self.param_desc[k]['prior'].transform(t_params[k])
				t_p_std[k] = t_params[k] / z_temp

			# Replace with something more elegant in future versions
			if printer is True:

				data = []

				for i in range(len(self.param_desc)-self.param_hide):
					if self.param_desc[i]['prior'].transform == np.array:
						data.append({'param_name': self.param_desc[i]['name'], 'param_value':round(self.param_desc[i]['prior'].transform(p.x[i]),4), 'param_std': round(t_p_std[i],4),'param_z': round(t_params[i]/float(t_p_std[i]),4),'param_p': round(tst.find_p_value(t_params[i]/float(t_p_std[i])),4),'ci': "(" + str(round(t_params[i] - t_p_std[i]*1.96,4)) + " | " + str(round(t_params[i] + t_p_std[i]*1.96,4)) + ")"})
					else:
						data.append({'param_name': self.param_desc[i]['name'], 'param_value':round(self.param_desc[i]['prior'].transform(p.x[i]),4)})						
				fmt = [
				    ('Parameter',       'param_name',   40),
				    ('Estimate',          'param_value',       10),
				    ('Standard Error', 'param_std', 15),
				    ('z',          'param_z',       10),
				    ('P>|z|',          'param_p',       10),
				    ('95% Confidence Interval',          'ci',       25)
				]

		except: # If Hessian is not invertible...
			print "Hessian not invertible! Consider a different model specification."
			print ""

			# Reset iHessian and standard errors, reflecting non-invertibility
			self.ihessian = []
			self.ses = []

			# Transform parameters
			for k in range(len(t_params)):
				t_params[k] = self.param_desc[k]['prior'].transform(t_params[k])

			if printer is True:

				data = []

				for i in range(len(self.param_desc)):
					data.append({'param_name': self.param_desc[i]['name'], 'param_value':round(self.param_desc[i]['prior'].transform(p.x[i]),4)})

				fmt = [
				    ('Parameter',       'param_name',   40),
				    ('Estimate',          'param_value',       10)
				]

		# Final printed output
		if printer is True:

			if self.model_type == 'ARIMA':
				self.model_name = "ARIMA(" + str(self.ar) + "," + str(self.integ) + "," + str(self.ma) + ") regression"
				self.cutoff = self.max_lag	
				self.data_length = self.data
			elif self.model_type == 'GAS':
				self.model_name = "GAS(" + str(self.ar) + "," + str(self.integ) + "," + str(self.sc) + ") regression"
				self.cutoff = self.max_lag	
				self.data_length = self.data				
			elif self.model_type == 'VAR':
				self.model_name = "VAR(" + str(self.lags) + ") regression"
				self.cutoff = self.lags			
				self.data_length = self.data[0]
			elif self.model_type in ['EGARCH', 'GARCH']:
				self.model_name = self.model_type + "(" + str(self.p) + "," + str(self.q) + ") regression"
				self.cutoff = self.max_lag		
				self.data_length = self.data	
			elif self.model_type == 'GPNARX':
				self.model_name = "GP-NARX(" + str(self.max_lag) + ") regression"
				self.cutoff = self.max_lag			
				self.data_length = self.data						

			print self.model_name 
			print "=================="
			if obj_type == self.likelihood:
				print "Method: MLE"
			elif obj_type == self.posterior:
				print "Method: MAP"
			print "Number of observations: " + str(len(self.data_length)-self.cutoff)
			print "Log Likelihood: " + str(round(-self.likelihood(p.x),4))
			if obj_type == self.posterior:
				print "Unnormalized Log Posterior: " + str(round(-self.posterior(p.x),4))
			print "AIC: " + str(round(2*len(p.x)+2*self.likelihood(p.x),4))
			print "BIC: " + str(round(2*self.likelihood(p.x) + len(p.x)*log(len(self.data_length)-self.cutoff),4))
			print ""
			print( op.TablePrinter(fmt, ul='=')(data) )

	def ols_fit(self,printer):
		""" Performs OLS

		Parameters
		----------

		printer : Boolean
			Whether to print results or not

		Returns
		----------
		None (stores parameters)
		"""

		self.params = self.create_B_direct().flatten()
		cov = self.estimate_cov()

		# Inelegant - needs refactoring
		for i in range(self.ylen):
			for k in range(self.ylen):
				if i == k:
					self.params = np.append(self.params,self.param_desc[len(self.params)]['prior'].itransform(cov[i,k]))
				elif i > k:
					self.params = np.append(self.params,self.param_desc[len(self.params)]['prior'].itransform(cov[i,k]))

		self.ihessian = self.estimator_cov('OLS')
		self.ses = np.abs(np.diag(self.ihessian))**0.5
		t_params = copy.deepcopy(self.params)			
		t_p_std = copy.deepcopy(self.ses) # vector for transformed standard errors
			
		# Create transformed variables
		for k in range(len(t_params)-self.param_hide):
			z_temp = (self.params[k]/float(self.ses[k]))
			t_params[k] = self.param_desc[k]['prior'].transform(t_params[k])
			t_p_std[k] = t_params[k] / z_temp

		# Replace with something more elegant in future versions
		if printer is True:

			data = []

			for i in range(len(self.param_desc)-self.param_hide):
				data.append({'param_name': self.param_desc[i]['name'], 'param_value':round(self.param_desc[i]['prior'].transform(self.params[i]),4), 'param_std': round(t_p_std[i],4),'param_z': round(t_params[i]/float(t_p_std[i]),4),'param_p': round(tst.find_p_value(t_params[i]/float(t_p_std[i])),4),'ci': "(" + str(round(t_params[i] - t_p_std[i]*1.96,4)) + " | " + str(round(t_params[i] + t_p_std[i]*1.96,4)) + ")"})

			fmt = [
			    ('Parameter',       'param_name',   40),
			    ('Estimate',          'param_value',       10),
			    ('Standard Error', 'param_std', 15),
			    ('z',          'param_z',       10),
			    ('P>|z|',          'param_p',       10),
			    ('95% Confidence Interval',          'ci',       25)
				]		

			if self.model_type == 'ARIMA':
				self.model_name = "ARIMA(" + str(self.ar) + "," + str(self.integ) + "," + str(self.ma) + ") regression"
				self.cutoff = max(self.ar,self.ma)
				self.data_length = self.data			
			elif self.model_type == 'VAR':
				self.model_name = "VAR(" + str(self.lags) + ") regression"
				self.cutoff = self.lags			
				self.data_length = self.data[0]

			print self.model_name 
			print "=================="
			print "Method: OLS"
			print "Number of observations: " + str(len(self.data_length)-self.cutoff)
			print "Log Likelihood: " + str(round(-self.likelihood(self.params),4))
			print "AIC: " + str(round(2*len(self.params)+2*self.likelihood(self.params),4))
			print "BIC: " + str(round(2*self.likelihood(self.params) + len(self.params)*log(len(self.data_length)-self.cutoff),4))
			print ""
			print( op.TablePrinter(fmt, ul='=')(data) )

	def mcmc_fit(self,scale=(2.38/sqrt(10000)),nsims=100000,printer=True,method="M-H",cov_matrix=None,**kwargs):
		""" Performs MCMC 

		Parameters
		----------
		scale : float
			Default starting scale

		nsims : int
			Number of simulations

		printer : Boolean
			Whether to print results or not

		method : str
			What type of MCMC

		cov_matrix: None or np.ndarray
			Can optionally provide a covariance matrix for M-H.

		Returns
		----------
		None (plots posteriors, stores parameters)
		"""

		figsize = kwargs.get('figsize',(15,15))

		self.fit(method='MAP',printer=False)
		
		if method == "M-H":
			chain, mean_est, median_est, upper_95_est, lower_95_est = ifr.metropolis_hastings(self.posterior,scale,nsims,self.params,cov_matrix=cov_matrix)
		else:
			raise Exception("Method not recognized!")

		self.params = np.asarray(mean_est)

		for k in range(len(chain)):
			if self.param_desc[k]['prior'].transform == np.exp:
				chain[k] = np.exp(chain[k])
				mean_est[k] = exp(mean_est[k])
				median_est[k] = exp(median_est[k])
				upper_95_est[k] = exp(upper_95_est[k])
				lower_95_est[k] = exp(lower_95_est[k])				
			elif self.param_desc[k]['prior'].transform == np.tanh: 
				chain[k] = np.tanh(chain[k])
				mean_est[k] = tanh(mean_est[k])
				median_est[k] = tanh(median_est[k])				
				upper_95_est[k] = tanh(upper_95_est[k])
				lower_95_est[k] = tanh(lower_95_est[k])	

		self.chains = chain

		if printer == True:

			data = []

			for i in range(len(self.param_desc)):
				data.append({'param_name': self.param_desc[i]['name'], 'param_mean':round(mean_est[i],4), 'param_median':round(median_est[i],4), 'ci': "(" + str(round(lower_95_est[i],4)) + " | " + str(round(upper_95_est[i],4)) + ")"})

		fmt = [
			('Parameter','param_name',20),
			('Median','param_median',10),
			('Mean', 'param_mean', 15),
			('95% Credibility Interval','ci',25)]

		if self.model_type == 'ARIMA':
			self.model_name = "ARIMA(" + str(self.ar) + "," + str(self.integ) + "," + str(self.ma) + ") regression"
			self.cutoff = max(self.ar,self.ma)
		elif self.model_type == 'GAS':
			self.model_name = "GAS(" + str(self.ar) + "," + str(self.integ) + "," + str(self.sc) + ") regression"
			self.cutoff = max(self.ar,self.sc)
		elif self.model_type == 'VAR':
			self.model_name = "VAR(" + str(self.lags) + ") regression"
			self.cutoff = self.lags						
		elif self.model_type in ['EGARCH', 'GARCH']:
			self.model_name = self.model_type + "(" + str(self.p) + "," + str(self.q) + ") regression"
			self.cutoff = max(self.p,self.q)
		elif self.model_type == 'GPNARX':
			self.model_name = "GP-NARX(" + str(self.ar) + ") regression"
			self.cutoff = self.max_lag	

		print self.model_name 
		print "=================="
		print "Method: Metropolis-Hastings"
		print "Number of simulations: " + str(nsims)
		print "Number of observations: " + str(len(self.data)-self.cutoff)
		print "Unnormalized Log Posterior: " + str(round(-self.posterior(mean_est),4))
		print ""
		print( op.TablePrinter(fmt, ul='=')(data) )

		fig = plt.figure(figsize=figsize)

		for j in range(len(self.params)):

			for k in range(4):
				iteration = j*4 + k + 1
				ax = fig.add_subplot(len(self.params),4,iteration)

				if iteration in range(1,len(self.params)*4 + 1,4):
					a = sns.distplot(chain[j], rug=False, hist=False)
					a.set_ylabel(self.param_desc[j]['name'])
					if iteration == 1:
						a.set_title('Density Estimate')
				elif iteration in range(2,len(self.params)*4 + 1,4):
					a = plt.plot(chain[j])
					if iteration == 2:
						plt.title('Trace Plot')
				elif iteration in range(3,len(self.params)*4 + 1,4): 
					plt.plot(np.cumsum(chain[j])/np.array(range(1,len(chain[j])+1)))
					if iteration == 3:
						plt.title('Cumulative Average')					
				elif iteration in range(4,len(self.params)*4 + 1,4):
					plt.bar(range(1,10),[cov.acf(chain[j],lag) for lag in range(1,10)])
					if iteration == 4:
						plt.title('ACF Plot')						
		sns.plt.show()		


	def plot_fit(self,**kwargs):
		""" Plots the fit of the model

		Returns
		----------
		None (plots data and the fit)
		"""

		figsize = kwargs.get('figsize',(10,7))

		if len(self.params) == 0:
			raise Exception("No parameters estimated!")
		else:
			if self.model_type == 'GAS':
				plt.figure(figsize=figsize)
				date_index = self.index[max(self.ar,self.sc):len(self.data)]
				mu, Y, scores = self.model(self.params)
				plt.plot(date_index,Y,label='Data')
				plt.plot(date_index,self.link(mu),label='Filter',c='black')
				plt.title(self.data_name)
				plt.legend(loc=2)	
			elif self.model_type == 'VAR':
				date_index = self.index[self.lags:len(self.data[0])]
				mu, Y = self.model(self.params)
				for series in range(len(Y)):
					plt.figure(figsize=figsize)
					plt.plot(date_index,Y[series],label='Data ' + str(series))
					plt.plot(date_index,mu[series],label='Filter' + str(series),c='black')	
					plt.title(self.data_name[series])
					plt.legend(loc=2)	
			elif self.model_type in ['GARCH','EGARCH']:
				plt.figure(figsize=figsize)
				date_index = self.index[max(self.p,self.q):len(self.data)]
				sigma2, Y, ___ = self.model(self.params)
				plt.plot(date_index,np.abs(Y),label=self.data_name + ' Absolute Values')
				if self.model_type == 'GARCH':
					plt.plot(date_index,np.power(sigma2,0.5),label='GARCH(' + str(self.p) + ',' + str(self.q) + ') std',c='black')
				elif self.model_type == 'EGARCH':
					plt.plot(date_index,np.exp(sigma2/2),label='EGARCH(' + str(self.p) + ',' + str(self.q) + ') std',c='black')					
				plt.title(self.data_name + " Volatility Plot")	
				plt.legend(loc=2)	
			elif self.model_type == 'GPNARX':
				self.pfit(self.params)		
			else:
				plt.figure(figsize=figsize)
				date_index = self.index[max(self.ar,self.ma):len(self.data)]
				mu, Y = self.model(self.params)
				plt.plot(date_index,Y,label='Data')
				plt.plot(date_index,mu,label='Filter',c='black')
				plt.title(self.data_name)
				plt.legend(loc=2)	
			plt.show()				

	def bbvi_fit(self,posterior,printer=True,step=0.001,iterations=30000):
		""" Performs Black Box Variational Inference

		Parameters
		----------
		posterior : method
			Hands bbvi_fit a posterior object

		printer : Boolean
			Whether to print results or not

		step : float
			Step size for RMSProp

		iterations: int
			How many iterations for BBVI

		Returns
		----------
		None (plots posteriors and stores parameters)
		"""

		# Starting parameters
		phi = np.zeros(self.param_no)

		# Starting values
		if self.model_type == 'GAS':
			if self.dist in ['Laplace','Normal']:
				phi[0] = np.mean(self.data)
			elif self.dist == 'Poisson':
				phi[0] = np.log(np.mean(self.data))
		elif self.model_type == 'VAR':
			phi = self.create_B_direct().flatten()
			cov = self.estimate_cov()

			# Inelegant - needs refactoring
			for i in range(self.ylen):
				for k in range(self.ylen):
					if i == k:
						phi = np.append(phi,self.param_desc[len(phi)]['prior'].itransform(cov[i,k]))
					elif i > k:
						phi = np.append(phi,self.param_desc[len(phi)]['prior'].itransform(cov[i,k]))			
		elif self.model_type == 'GARCH':
			phi = np.ones(self.param_no)*0.00001
			phi[0] = self.param_desc[0]['prior'].itransform(np.mean(np.power(self.data,2)))
		elif self.model_type == 'EGARCH':
			phi[0] = self.param_desc[0]['prior'].itransform(np.log(np.mean(np.power(self.data,2))))
		elif self.model_type == 'GPNARX':
			phi = np.ones(self.param_no)*0.0
			arma_start = arma.ARIMA(self.data,ar=self.ar,ma=0,integ=self.integ)
			arma_start.fit(printer=False)
			phi[0] = arma_start.params[len(arma_start.params)-1]

		self.params = phi

		# Starting values for approximate distribution
		for i in range(len(self.param_desc)):
			approx_dist = self.param_desc[i]['q']
			if isinstance(approx_dist, dst.Normal):
				self.param_desc[i]['q'].loc = self.params[i]
				if len(self.ses) == 0:
					self.param_desc[i]['q'].scale = -3.0
				else:
					self.param_desc[i]['q'].scale = log(self.ses[i])
		q_list = [k['q'] for k in self.param_desc]
		
		bbvi_obj = ifr.BBVI(posterior,q_list,12,step,iterations)
		q, q_params, q_ses = bbvi_obj.lambda_update()
		self.params = q_params

		for k in range(len(self.param_desc)):
			self.param_desc[k]['q'] = q[k]

		self.normal_posterior_sim(self.params,np.diag(np.exp(q_ses)),"Black Box Variational Inference",printer)

	def shift_dates(self,h):
		""" Auxiliary function for creating dates for forecasts

		Parameters
		----------
		h : int
			How many steps to forecast

		Returns
		----------
		A transformed date_index object
		"""		

		if self.model_type == 'VAR':
			date_index = self.index[self.max_lag:len(self.data[0])]
		else:
			date_index = self.index[self.max_lag:len(self.data)]

		for t in range(h):
			if self.data_type == 'pandas':
				date_index += pd.DateOffset(1)
			elif self.data_type == 'numpy':
				date_index.append(date_index[len(date_index)-1]+1)

		return date_index	

	def transform_parameters(self):
		""" Transforms parameters to actual scale by applying link function

		Returns
		----------
		Transformed parameters 
		"""		
		trans_params = copy.deepcopy(self.params)
		for k in range(len(self.params)):
			trans_params[k] = self.param_desc[k]['prior'].transform(self.params[k])	
		return trans_params


import scipy.stats as ss

def find_p_value(z):
	p_value = 0
	if z>=0:
		p_value += (1-ss.norm.cdf(z,loc=0,scale=1))
		p_value += (ss.norm.cdf(-z,loc=0,scale=1))
	else:
		p_value += (1-ss.norm.cdf(-z,loc=0,scale=1))
		p_value += (ss.norm.cdf(z,loc=0,scale=1))
	return p_value


# Credit to Hugh Bothwell from http://stackoverflow.com/questions/5084743/how-to-print-pretty-string-output-in-python
class TablePrinter(object):
    "Print a list of dicts as a table"
    def __init__(self, fmt, sep=' ', ul=None):
        """        
        @param fmt: list of tuple(heading, key, width)
                        heading: str, column label
                        key: dictionary key to value to print
                        width: int, column width in chars
        @param sep: string, separation between columns
        @param ul: string, character to underline column label, or None for no underlining
        """
        super(TablePrinter,self).__init__()
        self.fmt   = str(sep).join('{lb}{0}:{1}{rb}'.format(key, width, lb='{', rb='}') for heading,key,width in fmt)
        self.head  = {key:heading for heading,key,width in fmt}
        self.ul    = {key:str(ul)*width for heading,key,width in fmt} if ul else None
        self.width = {key:width for heading,key,width in fmt}

    def row(self, data):
        return self.fmt.format(**{ k:str(data.get(k,''))[:w] for k,w in self.width.iteritems() })

    def __call__(self, dataList):
        _r = self.row
        res = [_r(data) for data in dataList]
        res.insert(0, _r(self.head))
        if self.ul:
            res.insert(1, _r(self.ul))
        return '\n'.join(res)


import numpy as np
from math import sqrt
import matplotlib.pyplot as plt
import seaborn as sns

# Returns sample autocovariance of two np arrays (stationarity assumed)
def cov(x,lag=0):
	if lag >= x.shape[0]:
		raise Exception("Not enough observations to compute autocovariance")
	else:
		x1 = x[lag:x.shape[0]]
		x2 = x[0:(x.shape[0]-lag)]
		return np.sum((x1-np.mean(x))*(x2-np.mean(x)))/x1.shape[0]

# Returns sample autocorrelation function of an np array (stationarity assumed)
def acf(x,lag=0):
		x1 = x[lag:x.shape[0]]
		return cov(x,lag=lag)/cov(x,lag=0)

# Returns acf plot
def acf_plot(x,max_lag=10):
	plt.figure(figsize=(8,4))
	ax = plt.subplot(111)
	ax.set_axis_bgcolor((0.2,0.2,0.2))
	plt.bar(range(1,max_lag),[acf(x,lag) for lag in range(1,max_lag)])
	plt.xlabel("Lag")
	plt.ylabel("Autocorrelation")
	plt.title("ACF Plot")
	plt.axhline(0+2/sqrt(len(x)),linestyle="--",color='white')
	plt.axhline(0-2/sqrt(len(x)),linestyle="--",color='white')
	plt.show()


from math import exp

def lik_score(rv,mean,var,dist):
	if dist == "Laplace":
		if float(var*abs(rv-mean)) == 0:
			return 0
		else:
			return float(rv-mean)/float(var*abs(rv-mean))

	elif dist == "Normal":
		if (rv-mean) == 0:
			return 0
		else:
			return (rv-mean)

	elif dist == "Poisson":
		if float(mean) == 0:
			return 0
		else:
			return float(rv-mean)/float(mean)

	elif dist == "Exponential":
		return 1.0 - mean*rv

	elif dist == "Beta-t":
		"""
		return (((var+1.0)*(rv**2.0))/float(var*exp(mean) + (rv**2))) - 1.0
		"""
		
		try:
			return (((var+1.0)*(rv**2))/float(var*exp(mean) + (rv**2))) - 1.0
		except:
			return -1.0



import scipy.stats as ss
from math import exp
import matplotlib.mlab as mlab
import matplotlib.pyplot as plt
import numpy as np

class Normal(object):

	def __init__(self,loc,scale):
		self.loc = loc
		self.scale = scale
		self.param_no = 2

	def sim(self,size): 
		return ss.norm.rvs(loc=self.loc,scale=exp(self.scale),size=size)

	def loc_score(self,x):
		return (x-self.loc)/(exp(self.scale)**2)

	def scale_score(self,x):
		# For scale = exp(x)
		return exp(-2*self.scale)*(x-self.loc)**2 - 1 

	# Indexed scale - for BBVI code
	def score(self,x,index):
		if index == 0:
			return self.loc_score(x)
		elif index == 1:
			return self.scale_score(x)

	def return_param(self,index):
		if index == 0:
			return self.loc
		elif index == 1:
			return self.scale

	def change_param(self,index,value):
		if index == 0:
			self.loc = value
		elif index == 1:
			self.scale = value

	def logpdf(self,x):
		return ss.norm.logpdf(x,loc=self.loc,scale=exp(self.scale))

	def plot_pdf(self):
		x = np.linspace(self.loc-exp(self.scale)*3.5,self.loc+exp(self.scale)*3.5,100)
		plt.plot(x,mlab.normpdf(x,self.loc,exp(self.scale)))
		plt.show()



import numpy as np

class SquaredExponential(object):

	def __init__(self, X, l, tau):

		self.X = X.transpose()
		self.Xo = X
		self.l = l
		self.tau = tau

	def K(self):
		K = np.zeros((self.X.shape[0],self.X.shape[0]))

		for i in xrange(0,self.X.shape[0]):
			for j in xrange(0,self.X.shape[0]):
				# Due to Toeplitz structure
				if i == j:
					K[i,j] = self.tau
				else:
					K[i,j] = self.tau*np.exp(-0.5*np.sum(np.power(self.X[i] - self.X[j],2))/self.l**2)

		K = K + np.identity(K.shape[0])*(10**-10)

		return K

	def Kstar(self,Xstar):
		K = np.zeros((self.X.shape[0],Xstar.shape[0]))

		for i in xrange(0,self.X.shape[0]):
			for j in xrange(0,Xstar.shape[0]):
				# Due to Toeplitz structure
				if i == j:
					K[i,j] = self.tau
				else:
					K[i,j] = self.tau*np.exp(-0.5*np.sum(np.power(self.X[i] - Xstar[j],2))/self.l**2)

		return K

	def Kstarstar(self,Xstar):
		K = np.zeros((Xstar.shape[0],Xstar.shape[0]))

		for i in xrange(0,Xstar.shape[0]):
			for j in xrange(0,Xstar.shape[0]):
				# Due to Toeplitz structure
				if i == j:
					K[i,j] = self.tau
				else:
					K[i,j] = self.tau*np.exp(-0.5*np.sum(np.power(Xstar[i] - Xstar[j],2))/self.l**2)

		return K

class RationalQuadratic(object):

	def __init__(self, X, l, a, tau):

		self.X = X.transpose()
		self.Xo = X
		self.l = l
		self.a = a
		self.tau = tau

	def K(self):
		K = np.zeros((self.X.shape[0],self.X.shape[0]))

		for i in xrange(0,self.X.shape[0]):
			for j in xrange(0,self.X.shape[0]):
				K[i,j] = self.tau*(1 + 0.5*np.sum(np.power(self.X[i] - self.X[j],2))/((self.l**2)*self.a))**-self.a

		K = K + np.identity(K.shape[0])*(10**-10)

		return K

	def Kstar(self,Xstar):
		K = np.zeros((self.X.shape[0],Xstar.shape[0]))

		for i in xrange(0,self.X.shape[0]):
			for j in xrange(0,Xstar.shape[0]):
				K[i,j] = self.tau*(1 + 0.5*np.sum(np.power(self.X[i] - Xstar[j],2))/((self.l**2)*self.a))**-self.a

		return K

	def Kstarstar(self,Xstar):
		K = np.zeros((Xstar.shape[0],Xstar.shape[0]))

		for i in xrange(0,Xstar.shape[0]):
			for j in xrange(0,Xstar.shape[0]):
				K[i,j] = self.tau*(1 + 0.5*np.sum(np.power(Xstar[i] - Xstar[j],2))/((self.l**2)*self.a))**-self.a

		return K


class OrnsteinUhlenbeck(object):

	def __init__(self, X, l, tau):

		self.X = X.transpose()
		self.Xo = X
		self.l = l
		self.tau = tau

	def K(self):
		K = np.zeros((self.X.shape[0],self.X.shape[0]))

		for i in xrange(0,self.X.shape[0]):
			for j in xrange(0,self.X.shape[0]):
				# Due to Toeplitz structure
				if i == j:
					K[i,j] = self.tau
				else:
					K[i,j] = self.tau*np.exp(-np.sum(np.abs(self.X[i] - self.X[j]))/self.l)

		K = K + np.identity(K.shape[0])*(10**-10)

		return K

	def Kstar(self,Xstar):
		K = np.zeros((self.X.shape[0],Xstar.shape[0]))

		for i in xrange(0,self.X.shape[0]):
			for j in xrange(0,Xstar.shape[0]):
				# Due to Toeplitz structure
				if i == j:
					K[i,j] = self.tau
				else:
					K[i,j] = self.tau*np.exp(-np.sum(np.abs(self.X[i] - Xstar[j]))/self.l)

		return K

	def Kstarstar(self,Xstar):
		K = np.zeros((Xstar.shape[0],Xstar.shape[0]))

		for i in xrange(0,Xstar.shape[0]):
			for j in xrange(0,Xstar.shape[0]):
				# Due to Toeplitz structure
				if i == j:
					K[i,j] = self.tau
				else:
					K[i,j] = self.tau*np.exp(-np.sum(np.abs(Xstar[i] - Xstar[j]))/self.l)

		return K




from .. import inference as ifr
from .. import distributions as dst
from .. import output as op
from .. import tests as tst
from .. import tsm as tsm
import numpy as np
import pandas as pd
import scipy.stats as ss
from math import exp, sqrt, log, tanh
import copy
from scipy import optimize
import matplotlib.pyplot as plt
import seaborn as sns
import numdifftools as nd

class GARCH(tsm.TSM):
	""" Inherits time series methods from TSM class.

	**** GENERALIZED AUTOREGRESSIVE CONDITIONAL HETEROSCEDASTICITY (GARCH) MODELS ****

	Parameters
	----------
	data : pd.DataFrame or np.ndarray
		Field to specify the time series data that will be used.

	p : int
		Field to specify how many GARCH terms the model will have. Warning:
		higher-order lag specifications often fail to return for optimization
		methods of inference (MLE/MAP).

	q : int
		Field to specify how many ARCH terms the model will have. Warning:
		higher-order lag specifications often fail to return for optimization
		methods of inference (MLE/MAP).

	target : str (pd.DataFrame) or int (np.ndarray)
		Specifies which column name or array index to use. By default, first
		column/array will be selected as the dependent variable.
	"""

	def __init__(self,data,p,q,target=None):

		# Initialize TSM object
		tsm.TSM.__init__(self,'GARCH')

		# Parameters
		self.p = p
		self.q = q
		self.param_no = self.p + self.q + 1
		self.max_lag = max(self.p,self.q)
		self.hess_type = 'numerical'
		self.param_hide = 0 # Whether to cutoff variance parameters from results
		self.supported_methods = ["MLE","MAP","Laplace","M-H","BBVI"]
		self.default_method = "MLE"

		# Check pandas or numpy
		if isinstance(data, pd.DataFrame):
			self.index = data.index			
			if target is None:
				self.data = data.ix[:,0].values
				self.data_name = data.columns.values[0]
			else:
				self.data = data[target]
				self.data_name = target					
			self.data_type = 'pandas'
			print str(self.data_name) + " picked as target variable"
			print ""

		elif isinstance(data, np.ndarray):
			self.data_name = "Series"		
			self.data_type = 'numpy'	
			if any(isinstance(i, np.ndarray) for i in data):
				if target is None:
					self.data = data[0]
					self.index = range(len(data[0]))
				else:
					self.data = data[target]
					self.index = range(len(data[target]))
				print "Nested list " + str(target) + " chosen as target variable"
				print ""
			else:
				self.data = data
				self.index = range(len(data))

		else:
			raise Exception("The data input is not pandas or numpy compatible!")	

		self.param_desc.append({'name' : 'Constant', 'index': 0, 'prior': ifr.Normal(0,3,transform='exp'), 'q': dst.Normal(0,3)})		
		
		# ARCH terms e^2 (q)
		for j in range(1,self.q+1):
			self.param_desc.append({'name' : 'q(' + str(j) + ')', 'index': j, 'prior': ifr.Normal(0,0.5,transform=None), 'q': dst.Normal(0,3)})
		
		# GARCH terms sigma (p)
		for k in range(self.q+1,self.p+self.q+1):
			self.param_desc.append({'name' : 'p(' + str(k-self.q) + ')', 'index': k, 'prior': ifr.Normal(0,0.5,transform=None), 'q': dst.Normal(0,3)})


	def model(self,beta):
		""" Creates the structure of the model

		Parameters
		----------
		beta : np.ndarray
			Contains untransformed starting values for parameters

		Returns
		----------
		sigma2 : np.ndarray
			Contains the values for the conditional volatility series

		Y : np.ndarray
			Contains the length-adjusted time series (accounting for lags)

		eps : np.ndarray
			Contains the squared residuals (ARCH terms) for the time series
		"""

		xeps = np.power(self.data,2)
		Y = np.array(self.data[self.max_lag:len(self.data)])
		eps = np.power(Y,2)
		X = np.ones(len(Y))

		# Transform parameters
		parm = [self.param_desc[k]['prior'].transform(beta[k]) for k in range(len(beta))]

		# ARCH terms
		if self.q != 0:
			for i in range(self.q):	
				X = np.vstack((X,xeps[(self.max_lag-i-1):(len(xeps)-i-1)]))
			sigma2 = np.matmul(np.transpose(X),parm[0:len(parm)-self.p])
		else:
			sigma2 = np.transpose(X*parm[0])

		# GARCH terms
		if self.p != 0:
			for t in range(len(Y)):
				if t < self.max_lag:
					sigma2[t] = parm[0]/(1-np.sum(parm[(self.q+1):(self.q+self.p+1)]))
				elif t >= self.max_lag:
					for k in range(self.p):
						sigma2[t] += parm[1+self.q+k]*(sigma2[t-1-k])

		return sigma2, Y, eps

	# Returns negative log likelihood
	def likelihood(self,beta):
		""" Creates the negative log-likelihood of the model

		Parameters
		----------
		beta : np.ndarray
			Contains untransformed starting values for parameters

		Returns
		----------
		The negative logliklihood of the model
		"""		

		sigma2, Y, __ = self.model(beta)
		return -np.sum(ss.norm.logpdf(Y,loc=np.zeros(len(sigma2)),scale=np.power(sigma2,0.5)))

	def mean_prediction(self,sigma2,Y,scores,h,t_params):
		""" Creates a h-step ahead mean prediction

		Parameters
		----------
		sigma2 : np.ndarray
			The past predicted values

		Y : np.ndarray
			The past data

		scores : np.ndarray
			The past scores

		h : int
			How many steps ahead for the prediction

		t_params : np.ndarray
			A vector of (transformed) parameters

		Returns
		----------
		h-length vector of mean predictions
		"""		

		# Create arrays to iteratre over
		sigma2_exp = copy.deepcopy(sigma2)
		scores_exp = copy.deepcopy(scores)

		# Loop over h time periods			
		for t in range(h):
			new_value = t_params[0]

			if self.q != 0:
				for j in range(1,self.q+1):
					new_value += t_params[j]*sigma2_exp[len(sigma2_exp)-j]

			if self.p != 0:
				for k in range(1,self.p+1):
					new_value += t_params[k+self.q]*scores_exp[len(scores_exp)-k]

			sigma2_exp = np.append(sigma2_exp,[new_value]) # For indexing consistency
			scores_exp = np.append(scores_exp,[0]) # expectation of score is zero

		return sigma2_exp

	def sim_prediction(self,sigma2,Y,scores,h,t_params,simulations):
		""" Simulates a h-step ahead mean prediction

		Parameters
		----------
		sigma2 : np.ndarray
			The past predicted values

		Y : np.ndarray
			The past data

		scores : np.ndarray
			The past scores

		h : int
			How many steps ahead for the prediction

		t_params : np.ndarray
			A vector of (transformed) parameters

		simulations : int
			How many simulations to perform

		Returns
		----------
		Matrix of simulations
		"""		

		sim_vector = np.zeros([simulations,h])

		for n in range(simulations):
			# Create arrays to iteratre over		
			sigma2_exp = copy.deepcopy(sigma2)
			scores_exp = copy.deepcopy(scores)

			# Loop over h time periods			
			for t in range(h):
				new_value = t_params[0]

				if self.q != 0:
					for j in range(1,self.q+1):
						new_value += t_params[j]*sigma2_exp[len(sigma2_exp)-j]

				if self.p != 0:
					for k in range(1,self.p+1):
						new_value += t_params[k+self.q]*scores_exp[len(scores_exp)-k]

				sigma2_exp = np.append(sigma2_exp,[new_value]) # For indexing consistency
				scores_exp = np.append(scores_exp,scores[np.random.randint(len(scores))]) # expectation of score is zero

			sim_vector[n] = sigma2_exp[(len(sigma2_exp)-h):(len(sigma2_exp))]

		return np.transpose(sim_vector)

	def summarize_simulations(self,mean_values,sim_vector,date_index,h,past_values):
		""" Summarizes a simulation vector and a mean vector of predictions

		Parameters
		----------
		mean_values : np.ndarray
			Mean predictions for h-step ahead forecasts

		sim_vector : np.ndarray
			N simulation predictions for h-step ahead forecasts

		date_index : pd.DateIndex or np.ndarray
			Dates for the simulations

		h : int
			How many steps ahead are forecast

		past_values : int
			How many past observations to include in the forecast plot

		intervals : Boolean
			Would you like to show prediction intervals for the forecast?
		"""	

		error_bars = []
		for pre in range(5,100,5):
			error_bars.append(np.insert([np.percentile(i,pre) for i in sim_vector] - mean_values[(len(mean_values)-h):(len(mean_values))],0,0))
		forecasted_values = mean_values[(len(mean_values)-h-1):(len(mean_values))]
		plot_values = mean_values[len(mean_values)-h-past_values:len(mean_values)]
		plot_index = date_index[len(date_index)-h-past_values:len(date_index)]
		return error_bars, forecasted_values, plot_values, plot_index


	def predict(self,h=5,past_values=20,intervals=True,**kwargs):

		figsize = kwargs.get('figsize',(10,7))
		
		""" Makes forecast with the estimated model

		Parameters
		----------
		h : int (default : 5)
			How many steps ahead would you like to forecast?

		past_values : int (default : 20)
			How many past observations to show on the forecast graph?

		intervals : Boolean
			Would you like to show prediction intervals for the forecast?

		Returns
		----------
		- Plot of the forecast
		- Error bars, forecasted_values, plot_values, plot_index
		"""		

		if len(self.params) == 0:
			raise Exception("No parameters estimated!")
		else:

			# Retrieve data, dates and (transformed) parameters
			sigma2, Y, scores = self.model(self.params)			
			date_index = self.shift_dates(h)
			t_params = self.transform_parameters()

			# Get mean prediction and simulations (for errors)
			mean_values = self.mean_prediction(sigma2,Y,scores,h,t_params)
			sim_values = self.sim_prediction(sigma2,Y,scores,h,t_params,15000)
			error_bars, forecasted_values, plot_values, plot_index = self.summarize_simulations(mean_values,sim_values,date_index,h,past_values)

			plt.figure(figsize=figsize)
			if intervals == True:
				alpha =[0.15*i/float(100) for i in range(50,12,-2)]
				for count, pre in enumerate(error_bars):
					plt.fill_between(date_index[len(date_index)-h-1:len(date_index)], forecasted_values-pre, forecasted_values+pre,alpha=alpha[count])			
			plt.plot(plot_index,plot_values)
			plt.title("Forecast for " + self.data_name)
			plt.xlabel("Time")
			plt.ylabel(self.data_name)
			plt.show()

			self.predictions = {'error_bars' : error_bars, 'forecasted_values' : forecasted_values, 'plot_values' : plot_values, 'plot_index': plot_index}



import numpy as np
from math import exp
from scipy.stats import multivariate_normal

def metropolis_hastings(posterior,scale,nsims,initials,cov_matrix=None):
	phi = np.zeros([nsims,len(initials)])
	phi[0] = initials

	if cov_matrix is None:
		cov_matrix = np.identity(len(initials))

	# Initialize parameters for while loop
	acceptance = 1
	finish = 0

	while (acceptance < 0.234 or acceptance > 0.4) or finish == 0:
		
		# If acceptance is in range, proceed to sample, else continue tuning
		if not (acceptance < 0.234 or acceptance > 0.4):
			finish = 1
			print ""
			print "Tuning complete! Now sampling."
			sims_to_do = nsims
		else:
			sims_to_do = 5000 # For acceptance rate tuning
		
		# Holds data on acceptance rates and uniform random numbers
		a_rate = np.zeros([sims_to_do,1])
		crit = np.random.rand(sims_to_do,1)

		post = multivariate_normal(np.zeros(len(initials)),cov_matrix)
		rnums = post.rvs()*scale
		for k in range(1,sims_to_do): 
			rnums = np.vstack((rnums,post.rvs()*scale))

		old_lik = -posterior(phi[0])

		for i in range(1,sims_to_do):
			phi_prop = phi[i-1] + rnums[i]

			lik_rat = exp(-posterior(phi_prop) - old_lik)

			if crit[i] < lik_rat:
				phi[i] = phi_prop
				a_rate[i] = 1
				old_lik = -posterior(phi[i])
			else:
				phi[i] = phi[i-1]
				a_rate[i] = 0

		acceptance = sum(a_rate)/len(a_rate)
			
		# Acceptance rate tuning 

		if acceptance > 0.4:
			scale *= 1.3
		elif acceptance < 0.234 and acceptance > 0.1:
			scale *= (1/1.3)
		elif acceptance <= 0.1 and acceptance > 0.05:
			scale *= 0.4
		elif acceptance <= 0.05 and acceptance > 0.01:
			scale *= 0.2
		elif acceptance <= 0.01:
			scale *= 0.05

		print "Acceptance rate of Metropolis-Hastings is" + str(sum(a_rate)/len(a_rate))

	chain = np.array([phi[i][0] for i in range(len(phi))])
	for m in range(1,len(initials)):
		chain = np.vstack((chain,[phi[i][m] for i in range(len(phi))]))

	mean_est = [np.mean(np.array([phi[i][j] for i in range(len(phi))])) for j in range(len(initials))]
	median_est = [np.median(np.array([phi[i][j] for i in range(len(phi))])) for j in range(len(initials))]
	upper_95_est = [np.percentile(np.array([phi[i][j] for i in range(len(phi))]),95) for j in range(len(initials))]
	lower_95_est = [np.percentile(np.array([phi[i][j] for i in range(len(phi))]),5) for j in range(len(initials))]

	return chain, mean_est, median_est, upper_95_est, lower_95_est


from .. import inference as ifr
from .. import distributions as dist
from .. import output as op
from .. import tests as tst
from .. import tsm as tsm
import numpy as np
import pandas as pd
import scipy.stats as ss
from math import exp, sqrt, log, tanh, pi
import copy
from scipy import optimize
import matplotlib.pyplot as plt
import seaborn as sns
import numdifftools as nd
import datetime

class VAR(tsm.TSM):
	""" Inherits time series methods from TSM class.

	**** VECTOR AUTOREGRESSION (VAR) MODELS ****

	Parameters
	----------
	data : pd.DataFrame or np.ndarray
		Field to specify the time series data that will be used.

	lags : int
		Field to specify how many lag terms the model will have. 

	integ : int (default : 0)
		Specifies how many time to difference the dependent variables.

	target : str (pd.DataFrame) or int (np.ndarray)
		Specifies which column name or array index to use. By default, first
		column/array will be selected as the dependent variable.
	"""

	def __init__(self,data,lags,target=None,integ=0):

		# Initialize TSM object
		tsm.TSM.__init__(self,'VAR')

		# Parameters
		self.lags = lags
		self.int = integ
		self.max_lag = lags
		self.hess_type = 'numerical'
		self.supported_methods = ["OLS","MLE","MAP","Laplace","M-H","BBVI"]
		self.default_method = "OLS"

		# Check pandas or numpy
		if isinstance(data, pd.DataFrame):
			self.index = data.index		
			self.data = data.values
			self.data_name = data.columns.values
			self.data_type = 'pandas'
			print str(self.data_name) + " picked as target variables"
			print ""

		elif isinstance(data, np.ndarray):
			self.data_name = np.asarray(range(1,len(data)+1))	
			self.data_type = 'numpy'	
			self.data = data
			self.index = range(len(data[0]))

		else:
			raise Exception("The data input is not pandas or numpy compatible!")

		# Difference data
		X = np.transpose(self.data)
		for order in range(self.int):
			X = np.asarray([np.diff(i) for i in X])
			self.data_name = np.asarray(["Differenced " + str(i) for i in self.data_name])
		self.data = X		

		self.ylen = len(self.data_name)

		# Create VAR parameters
		for variable in range(self.ylen):
			self.param_desc.append({'name' : self.data_name[variable] + ' Constant', 'index': len(self.param_desc), 'prior': ifr.Normal(0,3,transform=None), 'q': dist.Normal(0,3)})		
			other_variables = np.delete(range(self.ylen), [variable])
			for lag_no in range(self.lags):
				self.param_desc.append({'name' : str(self.data_name[variable]) + ' AR(' + str(lag_no+1) + ')', 'index': len(self.param_desc), 'prior': ifr.Normal(0,0.5,transform=None), 'q': dist.Normal(0,3)})
				for other in other_variables:
					self.param_desc.append({'name' : str(self.data_name[other]) + ' to ' + str(self.data_name[variable]) + ' AR(' + str(lag_no+1) + ')', 'index': len(self.param_desc), 'prior': ifr.Normal(0,0.5,transform=None), 'q': dist.Normal(0,3)})

		# Variance prior
		for i in range(self.ylen):
			for k in range(self.ylen):
				if i == k:
					self.param_desc.append({'name' : 'Sigma' + str(self.data_name[i]),'index': len(self.param_desc),'prior': ifr.Uniform(transform='exp'), 'q': dist.Normal(0,3)})
				elif i > k:
					self.param_desc.append({'name' : 'Sigma' + str(self.data_name[i]) + ' to ' + str(self.data_name[k]),'index': len(self.param_desc), 'prior': ifr.Uniform(transform=None), 'q': dist.Normal(0,3)})

		# Other attributes
		self.param_hide = len(self.data)**2 - (len(self.data)**2 - len(self.data))/2 # Whether to cutoff variance parameters from results		
		self.param_no = len(self.param_desc)

	def model(self,beta):
		""" Creates the structure of the model

		Parameters
		----------
		beta : np.ndarray
			Contains untransformed starting values for parameters

		Returns
		----------
		mu : np.ndarray
			Contains the predicted values for the time series

		Y : np.ndarray
			Contains the length-adjusted time series (accounting for lags)
		"""		

		Y = np.array([reg[self.lags:len(reg)] for reg in self.data])

		# Transform parameters
		beta = [self.param_desc[k]['prior'].transform(beta[k]) for k in range(len(beta))]

		params = []
		col_length = 1 + self.ylen*self.lags
		for i in range(self.ylen):
			params.append(beta[(col_length*i): (col_length*(i+1))])

		mu = np.dot(np.array(params),self.create_Z(Y))

		return mu, Y

	def likelihood(self,beta):
		""" Creates the negative log-likelihood of the model

		Parameters
		----------
		beta : np.ndarray
			Contains untransformed starting values for parameters

		Returns
		----------
		The negative logliklihood of the model
		"""		

		mu, Y = self.model(beta)
		mu_t = np.transpose(mu)
		Y_t = np.transpose(Y)
		loglik = np.array([-ss.multivariate_normal.logpdf(mu_t[t], Y_t[t], self.create_cov(beta)) for t in range(len(Y[0]))])
		return np.sum(loglik)

	def create_Z(self,Y):
		""" Creates design matrix holding the lagged variables

		Parameters
		----------
		Y : np.ndarray
			The dependent variables Y

		Returns
		----------
		The design matrix Z
		"""	

		Z = np.ones(((self.ylen*self.lags +1),len(Y[0])))
		row_count = 1
		for lag in range(1,self.lags+1):
			for reg in range(len(Y)):
				Z[row_count,:] = self.data[reg][(self.lags-lag):len(self.data[reg])-lag]			
				row_count += 1
		return Z

	def create_B(self,Y):
		""" Creates coefficient matrix

		Parameters
		----------
		Y : np.ndarray
			The dependent variables Y

		Returns
		----------
		The coefficient matrix B
		"""			

		Z = self.create_Z(Y)
		return np.dot(np.dot(Y,np.transpose(Z)),np.linalg.inv(np.dot(Z,np.transpose(Z))))

	def create_B_direct(self):
		""" Creates coefficient matrix (calculates Y within - for OLS fitting)

		Returns
		----------
		The coefficient matrix B
		"""			

		Y = np.array([reg[self.lags:len(reg)] for reg in self.data])		
		Z = self.create_Z(Y)
		return np.dot(np.dot(Y,np.transpose(Z)),np.linalg.inv(np.dot(Z,np.transpose(Z))))

	def create_eps(self,Y):
		""" Creates the model residuals

		Parameters
		----------
		Y : np.ndarray
			The dependent variables Y

		Returns
		----------
		The model residuals
		"""			

		return (Y-np.dot(self.create_B(Y),self.create_Z(Y)))

	def estimate_cov(self):
		""" Creates OLS estimate of the covariance matrix

		Returns
		----------
		The OLS estimate of the covariance matrix
		"""			

		Y = np.array([reg[self.lags:len(reg)] for reg in self.data])		
		return (1.0/(len(Y[0])))*np.dot(self.create_eps(Y),np.transpose(self.create_eps(Y)))

	def create_cov(self,beta):
		""" Creates Covariance Matrix for a given Beta Vector

		Parameters
		----------
		beta : np.ndarray
			Contains untransformed starting values for parameters

		Returns
		----------
		A Covariance Matrix
		"""			

		cov_matrix = np.zeros((self.ylen,self.ylen))

		quick_count = 0
		for i in range(self.ylen):
			for k in range(self.ylen):
				if i >= k:
					index = self.ylen + self.lags*(self.ylen**2) + quick_count
					quick_count += 1
					cov_matrix[i,k] = self.param_desc[index]['prior'].transform(beta[index])

		return cov_matrix + np.transpose(np.tril(cov_matrix,k=-1))

	def estimator_cov(self,method):
		""" Wrapper function for covariance choice

		Parameters
		----------
		method : str
			Estimation method

		Returns
		----------
		A Covariance Matrix
		"""			
		Y = np.array([reg[self.lags:len(reg)] for reg in self.data])	
		Z = self.create_Z(Y)
		if method == 'OLS':
			sigma = self.estimate_cov()
		else:			
			sigma = self.create_cov(self.params)
		return np.kron(np.linalg.inv(np.dot(Z,np.transpose(Z))), sigma)

	def shock_create(self, h, shock_type, shock_index, shock_value, shock_dir, irf_intervals):
		""" Function creates shocks based on desired specification

		Parameters
		----------
		h : int
			How many steps ahead to forecast

		shock_type : None or str
			Type of shock; options include None, 'Cov' (simulate from covariance matrix), 'IRF' (impulse response shock)

		shock_index : int
			Which parameter to apply the shock to if using an IRF.

		shock_value : None or float
			If specified, applies a custom-sized impulse response shock.

		shock_dir : str
			Direction of the IRF shock. One of 'positive' or 'negative'.

		irf_intervals : Boolean
			Whether to have intervals for the IRF plot or not

		Returns
		----------
		A h-length list which contains np.ndarrays containing shocks for each variable
		"""		
		# Loop over the forecast period

		if shock_type is None:

			random = [np.zeros(self.ylen) for i in range(h)]

		elif shock_type == 'IRF':

			cov = self.create_cov(self.params)
			post = ss.multivariate_normal(np.zeros(self.ylen),cov)
			if irf_intervals is False:
				random = [np.zeros(self.ylen) for i in range(h)]
			else:
				random = [post.rvs() for i in range(h)]
				random[0] = np.zeros(self.ylen)

			if shock_value is None:
				if shock_dir=='positive':
					random[0][shock_index] = cov[shock_index,shock_index]**0.5
				elif shock_dir=='negative':
					random[0][shock_index] = -cov[shock_index,shock_index]**0.5
				else:
					raise ValueError("Unknown shock direction!")	
			else:
				random[0][shock_index] = shock_value		

		elif shock_type == 'Cov':
			
			cov = self.create_cov(self.params)
			post = ss.multivariate_normal(np.zeros(self.ylen),cov)
			random = [post.rvs() for i in range(h)]

		return random

	def forecast_mean(self,h,t_params,Y,shock_type=None,shock_index=0,shock_value=None,shock_dir='positive',irf_intervals=False):
		""" Function allows for mean prediction; also allows shock specification for simulations or impulse response effects

		Parameters
		----------
		h : int
			How many steps ahead to forecast

		t_params : np.ndarray
			Transformed parameter vector

		Y : np.ndarray
			Data for series that is being forecast

		shock_type : None or str
			Type of shock; options include None, 'Cov' (simulate from covariance matrix), 'IRF' (impulse response shock)

		shock_index : int
			Which parameter to apply the shock to if using an IRF.

		shock_value : None or float
			If specified, applies a custom-sized impulse response shock.

		shock_dir : str
			Direction of the IRF shock. One of 'positive' or 'negative'.

		irf_intervals : Boolean
			Whether to have intervals for the IRF plot or not

		Returns
		----------
		A vector of forecasted data
		"""			

		random = self.shock_create(h, shock_type, shock_index, shock_value, shock_dir,irf_intervals)
		exp = [Y[variable] for variable in range(self.ylen)]
		
		# Each forward projection
		for t in range(h):
			new_values = np.zeros(self.ylen)

			# Each variable
			for variable in range(self.ylen):
				index_ref = variable*(1+self.ylen*self.lags)
				new_values[variable] = t_params[index_ref] # constant

				# VAR(p) terms
				for lag in range(self.lags):
					for lagged_var in range(self.ylen):
						new_values[variable] += t_params[index_ref+lagged_var+(lag*self.ylen)+1]*exp[lagged_var][len(exp[lagged_var])-1-lag]
				
				# Random shock
				new_values[variable] += random[t][variable]

			# Add new values
			for variable in range(self.ylen):
				exp[variable] = np.append(exp[variable],new_values[variable])

		return np.array(exp)


	def irf(self,h=10,shock_index=0,shock_value=None,intervals=True,shock_dir='positive',cumulative=False,**kwargs):

		figsize = kwargs.get('figsize',(10,7))

		""" Function allows for mean prediction; also allows shock specification for simulations or impulse response effects

		Parameters
		----------
		h : int
			How many steps ahead to forecast

		shock_index : int
			Which parameter (index) to apply the shock to.

		shock_value : None or float
			If specified, applies a custom-sized impulse response shock.

		intervals : Boolean
			Whether to have intervals for the IRF plot or not

		shock_dir : str
			Direction of the IRF shock. One of 'positive' or 'negative'.

		cumulative : Boolean
			Whether to plot cumulative effect (if no -> plot effect at each step)

		Returns
		----------
		A plot of the impact response
		"""			

		if len(self.params) == 0:
			raise Exception("No parameters estimated!")
		else:		

			# Retrieve data, dates and (transformed) parameters
			mu, Y = self.model(self.params)	
			date_index = self.shift_dates(h)
			t_params = self.transform_parameters()

			# Get steady state values (hacky solution)
			ss_exps = self.forecast_mean(150,t_params,Y,None)
			ss_exps = np.array([i[len(i)-10:len(i)-1] for i in ss_exps])

			# Expectation
			exps = self.forecast_mean(h,t_params,ss_exps,'IRF',shock_index,None,shock_dir)

			if intervals is True:
				# Error bars
				sim_vector = np.array([np.zeros([10000,h]) for i in range(self.ylen)])
				for it in range(10000):
					exps_sim = self.forecast_mean(h,t_params,Y,'IRF',shock_index,None,shock_dir,True)
					for variable in range(self.ylen):
						sim_vector[variable][it,:] = exps_sim[variable][(len(exps_sim[variable])-h):(len(exps_sim[variable]))]

			for variable in range(len(exps)):

				exp_var = exps[variable][(len(exps[variable])-h-1):len(exps[variable])]
				exp_var = exp_var - ss_exps[variable][len(ss_exps[variable])-1] # demean

				if cumulative is True:
					exp_var = exp_var.cumsum()
					if intervals is True:
						sims = sim_vector[variable]
						sims = [i.cumsum() for i in sims]
						sims = np.transpose(sims)
				elif cumulative is False:
					if intervals is True:
						sims = np.transpose(sim_vector[variable])
						
				plt.figure(figsize=figsize)

				if intervals is True:
					error_bars, forecasted_values, plot_values, plot_index = self.summarize_simulations(exps[variable],sims,date_index,h,0)

					if cumulative is True:
						forecasted_values = (forecasted_values - ss_exps[variable][len(ss_exps[variable])-1]).cumsum()
					else:
						forecasted_values = (forecasted_values - ss_exps[variable][len(ss_exps[variable])-1])
					
					alpha =[0.15*i/float(100) for i in range(50,12,-2)]
					for count, pre in enumerate(error_bars):

						if cumulative is True:
							ebar = pre - ss_exps[variable][len(ss_exps[variable])-1]					
						else:
							ebar = pre - ss_exps[variable][len(ss_exps[variable])-1]

						plt.fill_between(range(1,len(forecasted_values)), forecasted_values[1:len(ebar)]-ebar[1:len(ebar)], forecasted_values[1:len(ebar)]+ebar[1:len(ebar)],alpha=alpha[count])			
					
				plt.plot(exp_var)
				plt.plot(np.zeros(len(exp_var)),alpha=0.3)	
				if shock_dir == 'positive':				
					plt.title("IR for " + self.data_name[variable] + " for +ve shock in " + self.data_name[shock_index])
				elif shock_dir == 'negative':
					plt.title("IR for " + self.data_name[variable] + " for -ve shock in " + self.data_name[shock_index])					
				plt.xlabel("Time")
				plt.ylabel(self.data_name[variable])
				plt.show()


	def summarize_simulations(self,mean_values,sim_vector,date_index,h,past_values):
		""" Summarizes a simulation vector and a mean vector of predictions

		Parameters
		----------
		mean_values : np.ndarray
			Mean predictions for h-step ahead forecasts

		sim_vector : np.ndarray
			N simulation predictions for h-step ahead forecasts

		date_index : pd.DateIndex or np.ndarray
			Dates for the simulations

		h : int
			How many steps ahead are forecast

		past_values : int
			How many past observations to include in the forecast plot

		intervals : Boolean
			Would you like to show prediction intervals for the forecast?
		"""				
		error_bars = []
		for pre in range(5,100,5):
			error_bars.append(np.insert([np.percentile(i,pre) for i in sim_vector] - mean_values[(len(mean_values)-h):(len(mean_values))],0,0))
		forecasted_values = mean_values[(len(mean_values)-h-1):(len(mean_values))]
		plot_values = mean_values[len(mean_values)-h-past_values:len(mean_values)]
		plot_index = date_index[len(date_index)-h-past_values:len(date_index)]
		return error_bars, forecasted_values, plot_values, plot_index


	def predict(self,h=5,past_values=20,intervals=True,**kwargs):

		figsize = kwargs.get('figsize',(10,7))

		""" Makes forecast with the estimated model

		Parameters
		----------
		h : int (default : 5)
			How many steps ahead would you like to forecast?

		past_values : int (default : 20)
			How many past observations to show on the forecast graph?

		intervals : Boolean
			Would you like to show prediction intervals for the forecast?

		Returns
		----------
		- Plot of the forecast
		- Error bars, forecasted_values, plot_values, plot_index
		"""				

		if len(self.params) == 0:
			raise Exception("No parameters estimated!")
		else:

			# Retrieve data, dates and (transformed) parameters
			mu, Y = self.model(self.params)	
			date_index = self.shift_dates(h)
			t_params = self.transform_parameters()

			# Expectation
			exps = self.forecast_mean(h,t_params,Y,None,None)

			# Simulation
			sim_vector = np.array([np.zeros([15000,h]) for i in range(self.ylen)])
			for it in range(15000):
				exps_sim = self.forecast_mean(h,t_params,Y,"Cov",None)
				for variable in range(self.ylen):
					sim_vector[variable][it,:] = exps_sim[variable][(len(exps_sim[variable])-h):(len(exps_sim[variable]))]

			for variable in range(len(exps)):
				test = np.transpose(sim_vector[variable])
				error_bars, forecasted_values, plot_values, plot_index = self.summarize_simulations(exps[variable],test,date_index,h,past_values)

				plt.figure(figsize=figsize)
				if intervals == True:
					alpha = [0.15*i/float(100) for i in range(50,12,-2)]
					for count, pre in enumerate(error_bars):
						plt.fill_between(date_index[len(date_index)-h-1:len(date_index)], forecasted_values-pre, forecasted_values+pre,alpha=alpha[count])			
				plt.plot(plot_index,plot_values)
				plt.title("Forecast for " + self.data_name[variable])
				plt.xlabel("Time")
				plt.ylabel(self.data_name[variable])
				plt.show()
				self.predictions.append({'error_bars' : error_bars, 'forecasted_values' : forecasted_values, 'plot_values' : plot_values, 'plot_index': plot_index})





from .. import inference as ifr
from .. import distributions as dst
from .. import output as op
from .. import tests as tst
from .. import tsm as tsm
import numpy as np
import pandas as pd
import scipy.stats as ss
from scipy.stats import multivariate_normal
from math import exp, sqrt, log, tanh, pi
import copy
from scipy import optimize
import matplotlib.pyplot as plt
import seaborn as sns
import numdifftools as nd
from kernels import *

class GPNARX(tsm.TSM):
	""" Inherits time series methods from TSM class.

	**** GAUSSIAN PROCESS NONLINEAR AUTOREGRESSIVE (GP-NARX) MODELS ****

	Parameters
	----------
	data : pd.DataFrame or np.ndarray
		Field to specify the time series data that will be used.

	ar : int
		Field to specify how many AR terms the model will have.

	kernel_type : str
		One of SE (SquaredExponential), OU (Ornstein-Uhlenbeck), RQ
		(RationalQuadratic). Defines kernel choice for GP-NAR.

	integ : int (default : 0)
		Specifies how many time to difference the time series.

	target : str (pd.DataFrame) or int (np.ndarray)
		Specifies which column name or array index to use. By default, first
		column/array will be selected as the dependent variable.
	"""

	def __init__(self,data,ar,kernel_type='SE',integ=0,target=None):

		# Initialize TSM object
		tsm.TSM.__init__(self,'GPNARX')

		# Parameters
		self.ar = ar
		self.integ = integ

		if kernel_type == 'ARD':
			self.param_no = 3 + self.ar
		elif kernel_type == 'RQ':
			self.param_no = 3 + 1
		else:
			self.param_no = 3

		self.max_lag = self.ar
		self.hess_type = 'numerical'
		self.param_hide = 0 # Whether to cutoff variance parameters from results
		self.supported_methods = ["MLE","MAP","Laplace","M-H","BBVI"]
		self.default_method = "MLE"		
		self.kernel_type = kernel_type
		self.denom = 1.0 # Redundant - remove in future version; used for normalization

		# Check pandas or numpy
		if isinstance(data, pd.DataFrame):
			self.index = data.index			
			if target is None:
				self.data = data.ix[:,0].values
				self.data_name = data.columns.values[0]
			else:
				self.data = (data[target] - np.mean(data[target])) / np.power(np.std(data[target]),self.denom)				
				self.data_name = target					
			self.data_type = 'pandas'
			print str(self.data_name) + " picked as target variable"
			print ""

		elif isinstance(data, np.ndarray):
			self.data_name = "Series"		
			self.data_type = 'numpy'	
			if any(isinstance(i, np.ndarray) for i in data):
				if target is None:
					self.data = data[0]			
					self.index = range(len(data[0]))
				else:
					self.data = data[target]			
					self.index = range(len(data[target]))
				print "Nested list " + str(target) + " chosen as target variable"
				print ""
			else:
				self.data = data					
				self.index = range(len(data))

		else:
			raise Exception("The data input is not pandas or numpy compatible!")

		# Difference data
		Y = self.data
		for order in range(self.integ):
			Y = np.diff(Y)
			self.data_name = "Differenced " + self.data_name
		self.index = self.index[self.integ:len(self.index)]

		# Apply normalization
		self.data_orig = Y		
		self.data = np.array(self.data_orig[self.max_lag:len(self.data_orig)]) # adjust for lags
		self.norm_mean = np.mean(self.data)
		self.norm_std = np.std(self.data)	
		self.data = (self.data - self.norm_mean) / np.power(self.norm_std,self.denom)

		# Define parameters

		self.param_desc.append({'name' : 'Noise Sigma','index': 0, 'prior': ifr.Uniform(transform='exp'), 'q': dst.Normal(0,3)})

		self.param_desc.append({'name' : 'l','index': 1, 'prior': ifr.Uniform(transform='exp'), 'q': dst.Normal(0,3)})

		if self.kernel_type == 'SE':
			self.kernel = SquaredExponential(self.X(),1,1)
		elif self.kernel_type == 'OU':
			self.kernel = OrnsteinUhlenbeck(self.X(),1,1)
		elif self.kernel_type == 'RQ':
			self.param_desc.append({'name' : 'alpha','index': len(self.param_desc), 'prior': ifr.Uniform(transform='exp'), 'q': dst.Normal(0,3)})
			self.kernel = RationalQuadratic(self.X(),1,1,1)

		self.param_desc.append({'name' : 'tau', 'index': len(self.param_desc), 'prior': ifr.Uniform(transform='exp'), 'q': dst.Normal(0,3)})


	def start_params(self,beta):
		""" Transforms parameters for use in kernels

		Parameters
		----------
		beta : np.ndarray
			Contains untransformed starting values for parameters

		Returns
		----------
		None (changes data in self.kernel)
		"""

		self.kernel.l = self.param_desc[1]['prior'].transform(beta[1])
		if self.kernel_type == 'RQ':
			self.kernel.a = self.param_desc[2]['prior'].transform(beta[2])			
			self.kernel.tau = self.param_desc[3]['prior'].transform(beta[3])	
		else:
			self.kernel.tau = self.param_desc[2]['prior'].transform(beta[2])			

	def X(self):
		""" Creates design matrix of variables to use in GP regression

		Returns
		----------
		The design matrix
		"""		

		# AR terms
		for i in xrange(0,self.ar):
			datapoint = self.data_orig[(self.max_lag-i-1):(len(self.data_orig)-i-1)]			
			if i==0:
				X = (datapoint - self.norm_mean)/np.power(self.norm_std,self.denom)
			else:
				X = np.vstack((X,(datapoint - self.norm_mean)/np.power(self.norm_std,self.denom)))
		return X

	def L(self,beta):
		""" Creates cholesky decomposition of covarianc ematrix

		Parameters
		----------
		beta : np.ndarray
			Contains untransformed starting values for parameters

		Returns
		----------
		The cholesky decomposition (L) of K
		"""	

		return np.linalg.cholesky(self.kernel.K()) + np.identity(self.kernel.K().shape[0])*self.param_desc[0]['prior'].transform(beta[0])

	def alpha(self,beta):
		""" Covariance-derived term to construct expectations. See Rasmussen & Williams.

		Parameters
		----------
		beta : np.ndarray
			Contains untransformed starting values for parameters

		Returns
		----------
		The alpha matrix/vector
		"""		

		L = self.L(beta)
		return np.linalg.solve(np.transpose(L),np.linalg.solve(L,np.transpose(self.data)))

	def E_fstar(self,beta):
		""" Expected values of the function given the covariance matrix and hyperparameters

		Parameters
		----------
		beta : np.ndarray
			Contains untransformed starting values for parameters

		Returns
		----------
		The expected values of the function
		"""		

		self.start_params(beta)
		return np.dot(np.transpose(self.kernel.K()),self.alpha(beta))

	def v(self,beta):
		""" Covariance term used for variance of function (currently not in use)

		Parameters
		----------
		beta : np.ndarray
			Contains untransformed starting values for parameters

		Returns
		----------
		Covariance term v.
		"""

		v = np.linalg.solve(self.L(beta),self.kernel.K())
		return v

	def var_fstar(self,beta):
		""" Covariance matrix for the estimated function

		Parameters
		----------
		beta : np.ndarray
			Contains untransformed starting values for parameters

		Returns
		----------
		Covariance matrix for the estimated function 
		"""		

		self.start_params(beta)
		return self.kernel.K() - np.dot(np.dot(np.transpose(self.kernel.K()),np.linalg.pinv(self.kernel.K() + np.identity(self.kernel.K().shape[0])*self.param_desc[0]['prior'].transform(beta[0]))),self.kernel.K())		

	def pfit(self,beta,intervals=True,**kwargs):
		""" Plots the fit of the Gaussian process model to the data

		Parameters
		----------
		beta : np.ndarray
			Contains untransformed starting values for parameters

		intervals : Boolean
			Whether to plot uncertainty intervals or not

		Returns
		----------
		None (plots the fit of the function)
		"""

		figsize = kwargs.get('figsize',(10,7))

		date_index = self.index[self.max_lag:len(self.index)]
		expectation = self.E_fstar(self.params)
		posterior = multivariate_normal(expectation,self.var_fstar(self.params),allow_singular=True)
		simulations = 500
		sim_vector = np.zeros([simulations,len(expectation)])

		for i in range(simulations):
			sim_vector[i] = posterior.rvs()

		error_bars = []
		for pre in range(5,100,5):
			error_bars.append([(np.percentile(i,pre)*self.norm_std + self.norm_mean) for i in sim_vector.transpose()] - (expectation*self.norm_std + self.norm_mean))

		plt.figure(figsize=figsize)	

		plt.subplot(2, 2, 1)
		plt.title(self.data_name + " Raw")	


		plt.plot(date_index,self.data*self.norm_std + self.norm_mean,'k')

		plt.subplot(2, 2, 2)

		plt.title(self.data_name + " Raw and Expected")	
		plt.plot(date_index,self.data*self.norm_std + self.norm_mean,'k',alpha=0.2)
		plt.plot(date_index,self.E_fstar(beta)*self.norm_std + self.norm_mean,'b')

		plt.subplot(2, 2, 3)

		plt.title(self.data_name + " Raw and Expected (with intervals)")	

		if intervals == True:
			alpha =[0.15*i/float(100) for i in range(50,12,-2)]
			for count, pre in enumerate(error_bars):
				plt.fill_between(date_index, (expectation*self.norm_std + self.norm_mean)-pre, (expectation*self.norm_std + self.norm_mean)+pre,alpha=alpha[count])		

		plt.plot(date_index,self.data*self.norm_std + self.norm_mean,'k',alpha=0.2)
		plt.plot(date_index,self.E_fstar(beta)*self.norm_std + self.norm_mean,'b')

		plt.subplot(2, 2, 4)

		plt.title("Expected " + self.data_name + " (with intervals)")	

		if intervals == True:
			alpha =[0.15*i/float(100) for i in range(50,12,-2)]
			for count, pre in enumerate(error_bars):
				plt.fill_between(date_index, (expectation*self.norm_std + self.norm_mean)-pre, (expectation*self.norm_std + self.norm_mean)+pre,alpha=alpha[count])		

		plt.plot(date_index,self.E_fstar(beta)*self.norm_std + self.norm_mean,'b')

		plt.show()

	def likelihood(self,beta):
		""" Creates the negative log marginal likelihood of the model

		Parameters
		----------
		beta : np.ndarray
			Contains untransformed starting values for parameters

		Returns
		----------
		The log marginal logliklihood of the model
		"""				
		self.start_params(beta)
		return -(-0.5*(np.dot(np.transpose(self.data),self.alpha(beta))) - np.sum(np.diag(self.L(beta))) - (len(self.data)/2)*log(2*pi))

	######### Functions for prediction #########

	def construct_predict(self,beta,h):	
		""" Creates h-step ahead forecasts for the Gaussian process

		Parameters
		----------
		beta : np.ndarray
			Contains untransformed starting values for parameters

		h: int
			How many steps ahead to forecast

		Returns
		----------
		- predictions
		- variance of predictions
		"""				

		Xstart = copy.deepcopy(self.X())
		Xstart = [i for i in Xstart]

		predictions = np.zeros(h)
		variances = np.zeros(h)

		for step in range(h):

			Xstar = []

			for lag in range(self.max_lag):

				if lag == 0:
					if step == 0:
						Xstar.append([self.data[len(self.data)-1]])
						Xstart[0] = np.append(Xstart[0],self.data[len(self.data)-1])
					else:
						Xstar.append([predictions[step-1]])
						Xstart[0] = np.append(Xstart[0],predictions[step-1])
				else:
					Xstar.append([Xstart[lag-1][len(Xstart[lag-1])-2]])
					Xstart[lag] = np.append(Xstart[lag],Xstart[lag-1][len(Xstart[lag-1])-2])

			Kstar = self.kernel.Kstar(np.transpose(np.array(Xstar)))

			predictions[step] = np.dot(np.transpose(Kstar),self.alpha(beta))
			variances[step] = self.kernel.Kstarstar(np.transpose(np.array(Xstar))) - np.dot(np.dot(np.transpose(self.kernel.Kstar(np.transpose(np.array(Xstar)))),np.linalg.pinv(self.kernel.K() + np.identity(self.kernel.K().shape[0])*self.param_desc[0]['prior'].transform(beta[0]))),self.kernel.Kstar(np.transpose(np.array(Xstar))))		
		return predictions, variances, predictions - 1.98*np.power(variances,0.5), predictions + 1.98*np.power(variances,0.5)

	def predict(self,h=5,past_values=20,intervals=True,**kwargs):

		""" Makes forecast with the estimated model

		Parameters
		----------
		h : int (default : 5)
			How many steps ahead would you like to forecast?

		past_values : int (default : 20)
			How many past observations to show on the forecast graph?

		intervals : Boolean
			Would you like to show 95% prediction intervals for the forecast?

		Returns
		----------
		- Plot of the forecast
		- Error bars, forecasted_values, plot_values, plot_index
		"""		

		figsize = kwargs.get('figsize',(10,7))

		if len(self.params) == 0:
			raise Exception("No parameters estimated!")
		else:

			predictions, variance, lower, upper = self.construct_predict(self.params,h)	
			full_predictions = np.append(self.data,predictions)
			full_lower = np.append(self.data,lower)
			full_upper = np.append(self.data,upper)
			date_index = self.shift_dates(h)
			lower = np.append(full_predictions[len(date_index)-h+1],lower)
			upper = np.append(full_predictions[len(date_index)-h+1],upper)

			# Plot values (how far to look back)
			plot_values = full_predictions[len(full_predictions)-h-past_values:len(full_predictions)]*self.norm_std + self.norm_mean
			plot_index = date_index[len(date_index)-h-past_values:len(date_index)]

			plt.figure(figsize=figsize)
			if intervals == True:
				plt.fill_between(date_index[len(date_index)-h-1:len(date_index)], lower*self.norm_std + self.norm_mean, upper*self.norm_std + self.norm_mean,alpha=0.5)			
			
			plt.plot(plot_index,plot_values)
			plt.title("Forecast for " + self.data_name)
			plt.xlabel("Time")
			plt.ylabel(self.data_name)
			plt.show()

			self.predictions = {'error_bars' : np.array([lower,upper]), 'forecasted_values' : predictions, 'plot_values' : plot_values, 'plot_index': plot_index}



import numpy as np

class SquaredExponential(object):

	def __init__(self, X, l, tau):

		self.X = X
		self.l = l
		self.tau = tau

	def K(self):
		K = np.zeros((X.shape[0],X.shape[0]))

		for i in xrange(0,self.X.shape[0]):
			for j in xrange(0,self.X.shape[0]):
				K[i,j] = self.tau*np.exp((X.transpose()[i] - X.transpose()[j])/self.l**2)

		return K


from distutils.core import setup
setup(
  name = 'pyflux',
  packages = ['pyflux','pyflux.arma','pyflux.distributions','pyflux.gas','pyflux.garch','pyflux.inference','pyflux.output','pyflux.tests','pyflux.var','pyflux.gpnarx'], 
  version = '0.1.8',
  description = 'A time-series analysis library for Python',
  author = 'Ross Taylor',
  author_email = 'rj-taylor@live.co.uk',
  url = 'https://github.com/rjt1990/pyflux', 
  download_url = 'https://github.com/rjt1990/pyflux/tarball/0.1.8', 
  keywords = ['time series','machine learning','bayesian statistics'],
  license = 'BSD',
  install_requires=['numpy', 'pandas', 'scipy', 'matplotlib', 'seaborn','numdifftools']
)


import numpy as np
import numdifftools as nd
import matplotlib.pyplot as plt
import matplotlib.mlab as mlab

# Let x be observations, z be latent variables, lambda free parameters
# of variational distribution q(z given lambda).

# We want to approximate p(z given x with free parameter lambda).

class BBVI(object):

	def __init__(self,neg_posterior,q,sims,step=0.001,iterations=30000):
		self.neg_posterior = neg_posterior
		self.q = q
		self.sims = sims
		self.iterations = iterations
		self.approx_param_no = np.array([i.param_no for i in self.q])
		self.step = step

	def draw_variables(self):
		z = self.q[0].sim(self.sims)
		for i in range(1,len(self.q)):
			z = np.vstack((z,self.q[i].sim(self.sims)))
		return z

	def create_logq(self,z):
		logq = 0
		for partial in range(len(self.q)):
			logq += self.q[partial].logpdf(z[partial])		
		return logq

	def grad_log_q(self,z):
		param_count = 0
		grad = np.zeros((np.sum(self.approx_param_no),self.sims))
		for core_param in range(len(self.q)):
			for approx_param in range(self.q[core_param].param_no):
				grad[param_count] = self.q[core_param].score(z[core_param],approx_param)		
				param_count += 1
		return grad

	def log_p(self,z):
		z_t = np.transpose(z)
		return np.array([-self.neg_posterior(i) for i in z_t])

	def log_q(self,z):
		z_t = np.transpose(z)
		logq = np.zeros(self.sims)
		for s in range(len(z[0])):
			logq[s] = self.create_logq(z_t[s])
		return logq

	def simple_gradient(self,z):
		lambda_grad = np.zeros(np.sum(self.approx_param_no))
		log_q = self.log_q(z)
		log_p = self.log_p(z)
		grad_log_q = self.grad_log_q(z)
		for lambda_i in range(np.sum(self.approx_param_no)):
			lambda_grad[lambda_i] = np.mean(grad_log_q[lambda_i]*(log_p-log_q))
		return lambda_grad

	def cv_gradient(self,z):
		gradient = np.zeros(np.sum(self.approx_param_no))
		log_q = self.log_q(z)
		log_p = self.log_p(z)
		grad_log_q = self.grad_log_q(z)
		for lambda_i in range(np.sum(self.approx_param_no)):
			alpha = np.cov(grad_log_q[lambda_i],grad_log_q[lambda_i]*(log_p-log_q))[0][1]/np.var(grad_log_q[lambda_i])				
			gradient[lambda_i] = np.mean(grad_log_q[lambda_i]*(log_p-log_q) - alpha*grad_log_q[lambda_i])
		return gradient

	def current_parameters(self):
		current = []
		for core_param in range(len(self.q)):
			for approx_param in range(self.q[core_param].param_no):
				current.append(self.q[core_param].return_param(approx_param))
		return np.array(current)

	def change_parameters(self,params):
		no_of_params = 0
		for core_param in range(len(self.q)):
			for approx_param in range(self.q[core_param].param_no):
				self.q[core_param].change_param(approx_param,params[no_of_params])
				no_of_params += 1

	def return_elbo(self,params):
		self.change_parameters(params)
		return np.mean(np.asarray([self.neg_posterior(i) for i in np.transpose(self.draw_variables())]))

	def print_progress(self,i,current_params):
		for split in range(1,11):
			if i == (round(self.iterations/10*split)-1):
				print str(split) + "0% done : ELBO is " + str(-self.neg_posterior(current_params)-self.create_logq(current_params))

	def lambda_update(self):
		Gjj = 0
		final_parameters = self.current_parameters()
		final_samples = 1
		for i in range(self.iterations):
			# Draw variables and gradients
			z = self.draw_variables()
			gradient = self.cv_gradient(z)
			gradient[np.isnan(gradient)] = 0
			# RMS prop
			Gjj = 0.99*Gjj + 0.01*(gradient**2)
			new_parameters = self.current_parameters() + self.step*(gradient/np.sqrt(Gjj))	
			self.change_parameters(new_parameters)

			# Print progress
			current_z = self.current_parameters()
			current_lambda = np.array([current_z[el] for el in range(len(current_z)) if el%2==0])		
			self.print_progress(i,current_lambda)

			# Construct final parameters using final 10% of samples
			if i > self.iterations-round(self.iterations/10):
				final_samples += 1
				final_parameters = final_parameters+current_z

		final_parameters = final_parameters/float(final_samples)
		final_means = np.array([final_parameters[el] for el in range(len(final_parameters)) if el%2==0])
		final_ses = np.array([final_parameters[el] for el in range(len(final_parameters)) if el%2!=0])
		print ""
		print "Final model ELBO is " + str(-self.neg_posterior(final_means)-self.create_logq(final_means))
		return self.q, final_means, final_ses



"""
The module Inference holds estimation procedures.
"""
from priors import Normal, InverseGamma, Uniform
from metropolis_hastings import metropolis_hastings
from norm_post_sim import norm_post_sim
from bbvi import BBVI




from arma import *
from var import *
from gas import *
from garch import *
from gpnarx import *
from inference import *
import distributions as dist
from covariances import *
from output import *
from tests import *
